package body MT19937 is
	
	pragma Suppress(All_Checks);
	use type Cardinal;
	
	MATRIX_A : constant Cardinal := 16#9908b0df#;
	UPPER_MASK : constant Cardinal := 2 ** (Cardinal'Size - 1);
	LOWER_MASK : constant Cardinal := not UPPER_MASK;
	
	function Random (Gen : Generator) return Cardinal is
		mag01 : constant array (Cardinal range 0..1) of Cardinal := (0, MATRIX_A);
		y : Cardinal;
	begin
		if Gen.Of_State = null then
			raise Constraint_Error;
		end if;
		declare
			S : State_Access := Gen.Of_State;
		begin
			if S.Condition >= N then
				for kk in 0..(N - M - 1) loop
					y := (S.Vector(kk) and UPPER_MASK) or (S.Vector(kk + 1) and LOWER_MASK);
					S.Vector(kk) := S.Vector(kk + M) xor Interfaces.Shift_Right(y, 1) xor mag01(y and 1);
				end loop;
				for kk in (N - M)..(N - 2) loop
					y := (S.Vector(kk) and UPPER_MASK) or (S.Vector(kk + 1) and LOWER_MASK);
					S.Vector(kk) := S.Vector(kk + (M - N)) xor Interfaces.Shift_Right(y, 1) xor mag01(y and 1);
				end loop;
				y := (S.Vector(N-1) and UPPER_MASK) or (S.Vector(0) and LOWER_MASK);
				S.Vector(N - 1) := S.Vector(M - 1) xor Interfaces.Shift_Right(y, 1) xor mag01(y and 1);
				S.Condition := 0;
			end if;
			y := S.Vector(Integer(S.Condition));
			S.Condition := S.Condition + 1;
		end;
		y := y xor Interfaces.Shift_Right(y, 11);
		y := y xor (Interfaces.Shift_Left(y, 7) and 16#9d2c5680#);
		y := y xor (Interfaces.Shift_Left(y, 15) and 16#efc60000#);
		y := y xor Interfaces.Shift_Right(y, 18);
		return y;
	end Random;
	
	procedure Reset (Gen : in out Generator) is
		use Ada.Calendar;
		Base_Time : constant Time := Time_Of(
			Year => Year_Number'First, 
			Month => Month_Number'First,
			Day => Day_Number'First);
		type Union_Initiator(F : Boolean := False) is record
			case F is
				when False => D : Duration;
				when True  => V : Cardinal_Vector(1..((Duration'Size + Cardinal'Size - 1) / Cardinal'Size));
			end case;
		end record;
		pragma Unchecked_Union(Union_Initiator);
		Initiator : Union_Initiator;
	begin
		Initiator.V := (others => 0);
		Initiator.D := Clock - Base_Time;
		Reset(Gen, Initiator.V);
	end Reset;
	
	procedure Reset (Gen : in out Generator; Initiator : Cardinal) is
	begin
		if Gen.Of_State = null then
			Gen.Of_State := new State;
		end if;
		declare
			S : State_Access := Gen.Of_State;
			V : Cardinal := Initiator;
		begin
			S.Vector(0) := V;
			for I in 1..(N - 1) loop
				V := 1812433253 * (V xor Interfaces.Shift_Right(V, 30)) + Cardinal(I);
				S.Vector(I) := V;
			end loop;
			S.Condition := N;
		end;
	end Reset;
	
	procedure Reset (Gen : in out Generator; Initiator : Cardinal_Vector) is
	begin
		Reset(Gen, 19650218);
		declare
			S : State_Access := Gen.Of_State;
			i : Integer := 1;
			j : Integer := 0;
		begin
			for K in reverse 1..Cardinal'Max(N, Initiator'Length) loop
				declare
					P : Cardinal := S.Vector(i - 1);
				begin
					S.Vector(i) := (S.Vector(i) xor ((P xor Interfaces.Shift_Right(P, 30)) * 1664525)) + Initiator(Initiator'First + j) + Cardinal(j); 
				end;
				i := i + 1;
				if i >= N then
					S.Vector(0) := S.Vector(N - 1); 
					i := 1;
				end if;
				j := (j + 1) rem Initiator'Length;
			end loop;
			for K in reverse 1..(N - 1) loop
				declare
					P : Cardinal := S.Vector(i - 1);
				begin
					S.Vector(i) := (S.Vector(i) xor ((P xor Interfaces.Shift_Right(P, 30)) * 1566083941)) - Cardinal(i); 
				end;
				i := i + 1;
				if i >= N then
					S.Vector(0) := S.Vector(N - 1);
					i := 1; 
				end if;
			end loop;
			S.Vector(0) := 16#80000000#; 
		end;
	end Reset;
	
	procedure Save  (Gen : Generator; To_State   : out State) is
	begin
		if Gen.Of_State = null then
			raise Constraint_Error;
		end if;
		To_State := Gen.Of_State.all;
	end Save;
	
	procedure Reset (Gen : in out Generator; From_State : State) is
	begin
		if Gen.Of_State = null then
			Gen.Of_State := new State;
		end if;
		Gen.Of_State.all := From_State;
	end Reset;
	
	function Image (Of_State    : State)  return String is
		Result : String(1..Max_Image_Width);
		procedure Hex(Position : Positive; Value : Cardinal) is
			Image : constant array (Cardinal range 0..15) of Character := "0123456789abcdef";
			V : Cardinal := Value;
		begin
			for I in 1..Image_Column loop
				Result(Position + Image_Column - I) := Image(V rem 16);
				V := Interfaces.Shift_Right(V, 4);
			end loop;
		end Hex;
	begin
		for I in N_Range loop
			Hex(Result'First + I * Image_Column, Of_State.Vector(I));
		end loop;
		Hex(Result'First + N * Image_Column, Of_State.Condition);
		return Result;
	end Image;
	
	function Value (Coded_State : String) return State is
		Result : State;
		procedure Hex(Position : Positive; Value : out Cardinal) is
		begin
			Value := 0;
			for I in 0..(Image_Column - 1) loop
				Value := Interfaces.Shift_Left(Value, 4);
				declare
					C : Character := Coded_State(Position + I);
				begin
					case C is
						when '0'..'9' => Value := Value or (Character'Pos(C) - Character'Pos('0'));
						when 'a'..'f' => Value := Value or (Character'Pos(C) - (Character'Pos('a') - 10));
						when 'A'..'F' => Value := Value or (Character'Pos(C) - (Character'Pos('A') - 10));
						when others   => raise Constraint_Error;
					end case;
				end;
			end loop;
		end Hex;
	begin
		if Coded_State'Length /= Max_Image_Width then
			raise Constraint_Error;
		end if;
		for I in N_Range loop
			Hex(Coded_State'First + I * Image_Column, Result.Vector(I));
		end loop;
		Hex(Coded_State'First + N * Image_Column, Result.Condition);
		return Result;
	end Value;
	
	function Random_0_To_1 (Gen : Generator) return Uniformly_Distributed is
	begin
		return Uniformly_Distributed'Base(Random(Gen)) * (1.0 / 4294967295.0);
	end Random_0_To_1;
	
	function Random_0_To_Less_1 (Gen : Generator) return Uniformly_Distributed is
	begin
		return Uniformly_Distributed'Base(Random(Gen)) * (1.0 / 4294967296.0);
	end Random_0_To_Less_1;
	
	function Random_Greater_0_To_Less_1 (Gen : Generator) return Uniformly_Distributed is
	begin
		return (Uniformly_Distributed'Base(Random(Gen)) + 0.5) * (1.0 / 4294967296.0);
	end Random_Greater_0_To_Less_1;
	
	function Random_53_0_To_1 (Gen : Generator) return Uniformly_Distributed is
		A : Cardinal := Interfaces.Shift_Right(Random(Gen), 5);
		B : Cardinal := Interfaces.Shift_Right(Random(Gen), 6);
	begin
		return (Uniformly_Distributed'Base(a) * 67108864.0 + Uniformly_Distributed'Base(b)) * (1.0 / 9007199254740992.0);
	end Random_53_0_To_1;
	
	package body Discrete_Random is
		function Random (Gen : Generator) return Result_Subtype is
			subtype Real is Uniformly_Distributed'Base;
			First_Position : constant Integer := Result_Subtype'Pos(Result_Subtype'First);
			Last_Position : constant Integer := Result_Subtype'Pos(Result_Subtype'Last);
			Width : constant Integer := Last_Position - First_Position + 1;
			Position : Integer := First_Position + Integer(Real'Floor(Real(Random_0_To_Less_1(Gen)) * Real(Width)));
		begin
			return Result_Subtype'Val(Position);
		end Random;
	end Discrete_Random;
	
	procedure Finalize (Object : in out Generator) is
	begin
		if Object.Of_State /= null then
			Free(Object.Of_State);
		end if;
	end Finalize;
	
end MT19937;
